#include "mpi.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

/*
extern double drand48();
*/

/* Pipeline version of the algorithm... */
/* we really need the velocities as well... */
typedef struct {
    double x, y, z;
    double mass;
    } Particle;
/* We use leapfrog for the time integration ... */
typedef struct {
    double xold, yold, zold;
    double fx, fy, fz;
    } ParticleV;


void InitParticles( Particle[], ParticleV [], unsigned int );
double ComputeForces( Particle [], ParticleV [], Particle [], unsigned int );
double ComputeNewPos( Particle [], ParticleV [], unsigned int, double, MPI_Comm );

#define MAX_PARTICLES_ON_CPU 3000
#define DEFAULT_TIME_STEPS 10

int main( int argc, char *argv[] )
{
    Particle  particles[MAX_PARTICLES_ON_CPU];   /* Particles on ALL nodes */
    ParticleV pv[MAX_PARTICLES_ON_CPU];          /* Particle velocity */
    Particle  sendbuf[MAX_PARTICLES_ON_CPU],     /* buffers */
	recvbuf[MAX_PARTICLES_ON_CPU];
    MPI_Request request[2];
    int rank, size, npart;
    unsigned int
                 totpart,              /* total number of particles */
                 cnt;                  /* number of times in loop */
    MPI_Datatype particletype;
    int         pipe, left, right, periodic;
    double      sim_t;                      /* Simulation time */
    double      time;                       /* Computation time */
    MPI_Comm    commring;
    MPI_Status  statuses[2];

    MPI_Init( &argc, &argv );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    MPI_Type_contiguous( 4, MPI_DOUBLE, &particletype );
    MPI_Type_commit( &particletype );

    /* Get the best ring in the topology */
    periodic = 1;
    MPI_Cart_create( MPI_COMM_WORLD, 1, &size, &periodic, 1, &commring );
    MPI_Cart_shift( commring, 0, 1, &left, &right );

    if (argc < 3) {
        cnt = DEFAULT_TIME_STEPS;
    }
    else {
        cnt = atoi(argv[2]);
    }
    if (argc < 2) {
    	totpart = MAX_PARTICLES_ON_CPU * size;
    }
    else {
    	totpart = atoi(argv[1]);
    }

    npart = totpart/size;
    if (totpart % size != 0) {
    	npart++;
    	totpart = npart * size;
    }

    if (npart > MAX_PARTICLES_ON_CPU) {
    	fprintf(stderr,"Error: Too much particles.\n");
    	MPI_Finalize();
    	return 1;
    }

/* Generate the initial values */
    memset (particles, 0 , npart*sizeof(Particle));
    InitParticles( particles, pv, npart);

    time = MPI_Wtime();
    sim_t = 0.0;
    unsigned int cnt_tmp = cnt;
    while (cnt_tmp--) {
		double max_f, max_f_seg;
		memcpy( sendbuf, particles, npart * sizeof(Particle) );
		max_f = 0.0;

		for (pipe=0; pipe<size; pipe++) {
		     if (pipe != size-1) {
		         MPI_Isend( sendbuf, npart, particletype, right, pipe, commring, &request[0] );
		         MPI_Irecv( recvbuf, npart, particletype, left,  pipe, commring, &request[1] );
		     }

		     /* Compute forces (2D only) */
		     max_f_seg = ComputeForces( particles, pv, sendbuf, npart );
		     if (max_f_seg > max_f) max_f = max_f_seg;
		     /* Push pipe */
		     if (pipe != size-1) {
		    	 MPI_Waitall( 2, request, statuses );
		    	 memcpy( sendbuf, recvbuf, npart * sizeof(Particle) );
		     }

		 }

		/* Once we have the forces, we compute the changes in position */
		sim_t += ComputeNewPos( particles, pv, npart, max_f, MPI_COMM_WORLD );

		/* We could do graphics here (move particles on the display) */
    } //while
    time = MPI_Wtime() - time;
    if (rank == 0) {
		printf( "Computed %d particles after %d steps in %f seconds\n", totpart, cnt, time );
    } //if
    MPI_Finalize();
    return 0;
}

void InitParticles( Particle particles[], ParticleV pv[], unsigned int npart )
{
    unsigned int i;
    for (i=0; i<npart; i++) {
		particles[i].x	  = drand48();
		particles[i].y	  = drand48();
		particles[i].z	  = drand48();
		particles[i].mass = 1.0;
		pv[i].xold	  = particles[i].x;
		pv[i].yold	  = particles[i].y;
		pv[i].zold	  = particles[i].z;
		pv[i].fx	  = 0;
		pv[i].fy	  = 0;
		pv[i].fz	  = 0;
    }
}

double ComputeForces( Particle particles[], ParticleV pv[], Particle sendbuf[], unsigned int npart )
{
  double max_f, rmin;
  int i, j;

  max_f = 0.0;
  for (i=0; i<npart; i++) {
    double xi, yi, rx, ry, mj, r, fx, fy;
    rmin = 100.0;
    xi   = particles[i].x;
    yi   = particles[i].y;
    fx   = 0.0;
    fy   = 0.0;
    for (j=0; j<npart; j++) {
      rx = xi - sendbuf[j].x;
      ry = yi - sendbuf[j].y;
      mj = sendbuf[j].mass;
      r  = rx * rx + ry * ry;
      /* ignore overlap and same particle */
      if (r == 0.0) continue;
      if (r < rmin) rmin = r;
      /* compute forces */
      r  = r * sqrt(r);
      fx -= mj * rx / r;
      fy -= mj * ry / r;
    }
    pv[i].fx += fx;
    pv[i].fy += fy;
    /* Compute a rough estimate of (1/m)|df / dx| */
    fx = sqrt(fx*fx + fy*fy)/rmin;
    if (fx > max_f) max_f = fx;
  }
  return max_f;
}

double ComputeNewPos( Particle particles[], ParticleV pv[], unsigned int npart,
				double max_f, MPI_Comm comm )
{
  int i;
  double      a0, a1, a2;
  static      double dt_old = 0.001, dt = 0.001;
  double      dt_est, dt_new;

  /* integration is a0 * x^+ + a1 * x + a2 * x^- = f / m */
  a0	 = 2.0 / (dt * (dt + dt_old));
  a2	 = 2.0 / (dt_old * (dt + dt_old));
  a1	 = -(a0 + a2);      /* also -2/(dt*dt_old) */

  for (i=0; i<npart; i++) {
    double xi, yi;
    /* Very, very simple leapfrog time integration.  We use a variable
       step version to simplify time-step control. */
    xi	           = particles[i].x;
    yi	           = particles[i].y;
    particles[i].x = (pv[i].fx - a1 * xi - a2 * pv[i].xold) / a0;
    particles[i].y = (pv[i].fy - a1 * yi - a2 * pv[i].yold) / a0;
    pv[i].xold     = xi;
    pv[i].yold     = yi;
    pv[i].fx       = 0;
    pv[i].fy       = 0;
  }

  /* Recompute a time step. Stability criteria is roughly
     2/sqrt(1/m |df/dx|) >= dt.  We leave a little room */
  dt_est = 1.0/sqrt(max_f);
  /* Set a minimum: */
  if (dt_est < 1.0e-6) dt_est = 1.0e-6;
  MPI_Allreduce( &dt_est, &dt_new, 1, MPI_DOUBLE, MPI_MIN, comm );
  /* Modify time step */
  if (dt_new < dt) {
    dt_old = dt;
    dt     = dt_new;
  }
  else if (dt_new > 4.0 * dt) {
    dt_old = dt;
    dt    *= 2.0;
  }

  return dt_old;
}
