/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <taag_basic.h>
#include <taag_errno.h>
#include <taag_xml.h>
#include <xml_parser.h>

static int _hw_initialized = TAAG_FALSE;
static int dumprsl =TAAG_FALSE;
static FILE *f;
static char fileName[TAAG_MAX_FILENAME_STRING];
static char executable[TAAG_MAX_FILENAME_STRING];
static char appId[TAAG_MAX_FILENAME_STRING];
static char command[TAAG_MAX_FILENAME_STRING];
static char arguments[TAAG_MAX_FILENAME_STRING];
static char xmlFileName[TAAG_MAX_FILENAME_STRING];

static unsigned int nrOfLANs = 0;
static unsigned int nrOfGrps = 0;
static int *nrOfProcsOnHosts;
static unsigned int *nrOfProcsInGrps;
static char **hostNames;
static char **directories;

//nr of all host contained by last element
//nr of hosts on i. LAN = i. element - (i-1). element
//hosts of i. LAN: (i-1). element  to i. element -1
//
//Advantage of this representation that we do not have to store
//the rank of the hosts;
static int *hostsOnLAN;

static int _getNrOfHost() {
	return(*(hostsOnLAN + nrOfLANs -1));
}

static int _getNrOfProcs(unsigned int nrOfGrps, unsigned int *procsInGrps) {
	unsigned int sum = 0;
	for (int i = 0; i < nrOfGrps; i++) {
		sum += *(procsInGrps+i);
	}
	return(sum);
}

int main(int argc, char **argv)
{
	int indexOfXMLArgument = 1;
	char *tmp;


	if (argc == 1) {
		fprintf(stderr, "Few arguments. Type \"taagrun -help\" for more details.\n");
		return(TAAG_ERR_COUNT);
	}
	if (!strcmp(argv[1],"-help")) {
		printf("taagrun [taagrun_options...] <XML file> [program arguments...]\n\n");
		printf("\ttaagrun options:\n");
		printf("\t\t-help    \tThis help.\n");
		printf("\t\t-dumprsl \tDisplays only the generated RSL file\n");
		printf("\t\t         \ton the standard output\n");
		printf("\t\t         \t(but it does not start the application\n");
		printf("\t\t         \ton the grid).\n");
		printf("\n");
		return(TAAG_SUCCESS);
	}
	if (!strcmp(argv[1],"-dumprsl")) {
		if (argc < 3) {
			fprintf(stderr, "Few arguments XML file name missing.\n");
			return(TAAG_ERR_COUNT);
		}
		indexOfXMLArgument = 2;
		dumprsl = TAAG_TRUE;
	}

	tmp = strrchr(argv[indexOfXMLArgument], '/');
	if (tmp == NULL) {
		strcpy(xmlFileName, argv[indexOfXMLArgument]);
	}
	else {
		strcpy(xmlFileName, tmp+1);
	}
	sprintf(arguments,"/tmp/%s", xmlFileName);
	for (int i = indexOfXMLArgument+1; i < argc; i++) {
		strcat(arguments, "\" \"");
		strcat(arguments, argv[i]);
	}

	if (_taag_xml_openFile(argv[indexOfXMLArgument])) {
		fprintf(stderr,"The XML file \"%s\" does not exist.\n",argv[indexOfXMLArgument]);
		return(TAAG_ERR_XML);
	}
	if (_taag_xml_validate()) return(TAAG_ERR_XML);

	_taag_xml_getAppId(appId);
	_taag_xml_getExecutable(executable);
	if (dumprsl) {
		f = stdout;
	}
	else {
		sprintf(fileName,"/tmp/%s.rsl", appId);
		f = fopen (fileName, "w");
	}

	//check wheather topology is specified
	_taag_xml_isTopologyGiven(&_hw_initialized);
	if (_hw_initialized == TAAG_FALSE) {
		//***** get number of groups
		_taag_xml_getNumberOfGroups(&nrOfGrps);

		//***** get sizes of all groups
		nrOfProcsInGrps = (unsigned int *)calloc(nrOfGrps, sizeof(unsigned int));
		_taag_xml_getGroupSizes(nrOfGrps, nrOfProcsInGrps);

		fprintf(stderr, "Warning: Topology is not specified in the XML-based description.\n");
		fprintf(stderr, "Executable is searched in the current directory and then\n");
		fprintf(stderr, "X processes are started on the localhost...\n\n");

		fprintf(f,"+\n");
		fprintf(f,"( &(resourceManagerContact=\"localhost\")\n");
		fprintf(f,"   (count=%d)\n", _getNrOfProcs(nrOfGrps, nrOfProcsInGrps));
		fprintf(f,"   (label=\"subjob 0\")\n");
		fprintf(f,"   (environment=(GLOBUS_DUROC_SUBJOB_INDEX 0)\n");
		fprintf(f,"                (GLOBUS_LAN_ID 0)\n");
		fprintf(f,"                (LD_LIBRARY_PATH $(GLOBUS_LOCATION)/lib/))\n");
		fprintf(f,"   (arguments=\"%s\")\n",arguments);
		fprintf(f,"   (directory=%s)\n", getenv("PWD"));
		fprintf(f,"   (executable=%s/%s)\n", getenv("PWD"), executable);
		fprintf(f,")\n");

		if (!dumprsl) {
			sprintf(command, "cp %s /tmp/%s", argv[indexOfXMLArgument], xmlFileName);
			system(command);
		}
		free(nrOfProcsInGrps);

	}
	else {
		_taag_xml_getNumberOfLANs(&nrOfLANs);

		//allocation for data structures
		hostsOnLAN = (int *)calloc(nrOfLANs, sizeof(int));
		_taag_xml_getHosts(nrOfLANs, hostsOnLAN);

		hostNames = (char **)calloc(_getNrOfHost(), sizeof(char *));
		directories = (char **)calloc(_getNrOfHost(), sizeof(char *));
		fprintf(f,"+\n");
		for (int i = 0; i < _getNrOfHost(); i++) {
			*(hostNames+i) = (char *)calloc(TAAG_MAX_HOSTNAME_STRING, sizeof(char));
			*(directories+i) = (char *)calloc(TAAG_MAX_HOSTNAME_STRING, sizeof(char));
		}
		_taag_xml_getHostnames(_getNrOfHost(), hostNames);
		_taag_xml_getDirectories(_getNrOfHost(), directories);

		nrOfProcsOnHosts = (int *)calloc(_getNrOfHost(), sizeof(int));
		memset(nrOfProcsOnHosts, 0, _getNrOfHost()*sizeof(int));
		_taag_xml_getNumberOfProcsOnHosts(_getNrOfHost(), nrOfProcsOnHosts);

		//g_location = getenv("GLOBUS_LOCATION"); //BUG!!! This is just the local GLOBUS_PATH
		//sprintf(ldLibrary,"%s/lib/",g_location);
		int lanId = 0;
		for (int hostId = 0; hostId < _getNrOfHost(); hostId++) {
			if (hostId == *(hostsOnLAN+lanId)) lanId++;
			fprintf(f,"( &(resourceManagerContact=\"%s\")\n", *(hostNames+hostId));
			fprintf(f,"   (count=%d)\n", *(nrOfProcsOnHosts+hostId));
			fprintf(f,"   (label=\"subjob %d\")\n", hostId);
			fprintf(f,"   (environment=(GLOBUS_DUROC_SUBJOB_INDEX %d)\n", hostId);
			fprintf(f,"                (GLOBUS_LAN_ID %d)\n",lanId);
			fprintf(f,"                (LD_LIBRARY_PATH $(GLOBUS_LOCATION)/lib/))\n");
			fprintf(f,"   (arguments=\"%s\")\n",arguments);
			fprintf(f,"   (directory=%s)\n", *(directories+hostId));
			fprintf(f,"   (executable=%s/%s)\n", *(directories+hostId), executable);
			fprintf(f,")\n");

			if (!dumprsl) {
				if (*argv[indexOfXMLArgument]=='/') {
					sprintf(command, "globus-url-copy file://%s gsiftp://%s/tmp/%s", argv[indexOfXMLArgument], *(hostNames+hostId), xmlFileName);
				}
				else {
					sprintf(command, "globus-url-copy file://%s/%s gsiftp://%s/tmp/%s", getenv("PWD"), argv[indexOfXMLArgument], *(hostNames+hostId), xmlFileName);
				}
				system(command);
			}
		} //for
		//free data structures
		for (int i = 0; i < _getNrOfHost(); i++) {
			tmp = *(hostNames+i);
			free(tmp);
			tmp = *(directories+i);
			free(tmp);
		} //for
		free(hostNames);
		free(directories);
		free(nrOfProcsOnHosts);
		free(hostsOnLAN);
	} //else

	//free xml parser
	_taag_xml_free();

	if (!dumprsl) {
		fclose(f);
		sprintf(command,"mpirun -globusrsl %s", fileName);
		system(command);
	}
}
