/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2010 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <map>

#include <taag_basic.h>
#include <taag_xml.h>

#include "LatencyMatrix.h"
#include "ProcsDistribution.h"
#include "GroupData.h"
#include "HostData.h"
#include "ListOfHosts.h"

/*
 *To compile this file using gcc you can type
 *gcc `xml2-config --cflags --libs` -o data2XMLmapping data2XMLmapping.c
 */

static char appId[TAAG_MAX_FILENAME_STRING];
static char dir[TAAG_MAX_FILENAME_STRING];
static char type[25];

static xmlDocPtr doc = NULL;       /* document pointer */
static xmlNodePtr root_node = NULL;
static xmlNodePtr graph = NULL;
static xmlNodePtr topology = NULL;

static map<int, int> lowestProcessIndexOnHosts;
static map<int, void *> LANs; //the addresses of 1. level latency clusters
static void *lanOfRoot; //address of lan (1. level latency cluster) which contains the root process
static int hostIdOfRoot; //id of host which contains the root process

using namespace std;

int _create_xml_mapping_heading(char *executable, char *schema, char *directory, time_t time) {
	xmlDtdPtr dtd = NULL;       /* DTD pointer */
	struct tm *time_st;
	char s_time[TAAG_MAX_HOSTNAME_STRING];

    LIBXML_TEST_VERSION;
    strcpy(type, schema);
    strcpy(dir, directory);
    time_st = localtime(&time);

    sprintf(s_time, "%d", 1900+time_st->tm_year);
    sprintf(s_time, "%s%s%d", s_time, (time_st->tm_mon+1<10) ? "0" : "", time_st->tm_mon+1);
    sprintf(s_time, "%s%s%d", s_time, (time_st->tm_mday<10) ? "0" : "", time_st->tm_mday);
    sprintf(s_time, "%s%s%d", s_time, (time_st->tm_hour<10) ? "0" : "", time_st->tm_hour);
    sprintf(s_time, "%s%s%d", s_time, (time_st->tm_min<10) ? "0" : "", time_st->tm_min);
    sprintf(s_time, "%s%s%d", s_time, (time_st->tm_sec<10) ? "0" : "", time_st->tm_sec);

    sprintf(appId,"%s_%s",executable, s_time);

    doc = xmlNewDoc(BAD_CAST "1.0");
    root_node = xmlNewNode(NULL, BAD_CAST TAAG_XML_ROOT);
    xmlDocSetRootElement(doc, root_node);

    /*
     * Creates a DTD declaration. Isn't mandatory.
     */
    dtd = xmlCreateIntSubset(doc, BAD_CAST TAAG_XML_ROOT, NULL, BAD_CAST "../../etc/taagXML.dtd");

	xmlNewChild(root_node, NULL, BAD_CAST TAAG_XML_APP_NAME, BAD_CAST executable);
	xmlNewChild(root_node, NULL, BAD_CAST TAAG_XML_APP_ID, BAD_CAST appId);
	xmlNewChild(root_node, NULL, BAD_CAST TAAG_XML_TIME_STAMP, BAD_CAST s_time);
	//graph = xmlNewNode(NULL, BAD_CAST TAAG_XML_GRAPH);
	//xmlAddChild(root_node, graph);

	return 0;
}

int _convert_mapping_to_XML(ProcsDistribution *procsDist, ListOfHosts *hostList) {
	int processIdCounter = 0;
	xmlNodePtr node = NULL;
	char s_data[TAAG_MAX_HOSTNAME_STRING];
	int nrOfGrps = 0;

	graph = xmlNewNode(NULL, BAD_CAST TAAG_XML_GRAPH);
	xmlNewChild(graph, NULL, BAD_CAST TAAG_XML_GRAPH_TYPE, BAD_CAST type);
	topology = xmlNewNode(NULL, BAD_CAST TAAG_XML_TOPOLOGY);

	nrOfGrps = procsDist->getNrOfGroups();
	if (nrOfGrps==0) return -1;

	GroupData *rootGrp = procsDist->getGroupByIndex(0);
	lanOfRoot = rootGrp->getMappingTicketByIndex(0)->getHost()->getAddressOfLANCluster();
	hostIdOfRoot = rootGrp->getMappingTicketByIndex(0)->getHost()->getHostId();

	for (int id=0; id<nrOfGrps; id++) {
		GroupData *grp = procsDist->getGroupByIndex(id);
		for (int j = 0; j < grp->getNrOfLoadTickets(); j++) {
			//initializing process id counters of hosts
			lowestProcessIndexOnHosts[grp->getMappingTicketByIndex(j)->getHost()->getHostId()]=0;

			//counting the number of processes scheduled to a host
			grp->getMappingTicketByIndex(j)->getHost()->increaseLoad(grp->getMappingTicketByIndex(j)->getNrOfProcs());

			//initializing LAN addresses (1. level latency cluster)
			LANs[grp->getMappingTicketByIndex(j)->getHost()->getHostId()]=grp->getMappingTicketByIndex(j)->getHost()->getAddressOfLANCluster();
		}
	}

	//******LANs*****
	int lan_counter = 0;

	//******BEGIN: Rank of root process must be 0 in case of trees*************************
	xmlNodePtr lan_node = xmlNewNode(NULL, BAD_CAST TAAG_XML_LAN);
	sprintf(s_data, "%d", lan_counter++);
	xmlNewProp(lan_node, BAD_CAST TAAG_XML_ATTR_ID, BAD_CAST s_data);

	LANs[hostIdOfRoot]=NULL;
	lowestProcessIndexOnHosts[hostIdOfRoot]=processIdCounter;
	HostData *host = hostList->getHostByIndex(hostIdOfRoot);

	node = xmlNewNode(NULL, BAD_CAST TAAG_XML_HOST);
	sprintf(s_data, "%d", host->getHostId());
	xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_ID, BAD_CAST s_data);
	xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_NAME, BAD_CAST host->getHostname());
	sprintf(s_data, "%d", host->getCPUcount());
	xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_CPUS, BAD_CAST s_data);
	xmlNewChild(node, NULL, BAD_CAST TAAG_XML_DIRECTORY, BAD_CAST dir);
	for (int i=0; i<host->getLoad(); i++) {
		sprintf(s_data, "%d", processIdCounter+i);
		xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_PROCS, BAD_CAST s_data);
	}
	processIdCounter+=host->getLoad();
	xmlAddChild(lan_node, node);

	//*****Hosts*****
	map<int, int>::iterator iter;
	for (iter=lowestProcessIndexOnHosts.begin(); iter != lowestProcessIndexOnHosts.end(); ++iter) {
		if (LANs[iter->first]==lanOfRoot) {
			LANs[iter->first]=NULL;
			lowestProcessIndexOnHosts[iter->first]=processIdCounter;
			HostData *host = hostList->getHostByIndex(iter->first);

			node = xmlNewNode(NULL, BAD_CAST TAAG_XML_HOST);
			sprintf(s_data, "%d", host->getHostId());
			xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_ID, BAD_CAST s_data);
			xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_NAME, BAD_CAST host->getHostname());
			sprintf(s_data, "%d", host->getCPUcount());
			xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_CPUS, BAD_CAST s_data);
			xmlNewChild(node, NULL, BAD_CAST TAAG_XML_DIRECTORY, BAD_CAST dir);
			for (int i=0; i<host->getLoad(); i++) {
				sprintf(s_data, "%d", processIdCounter+i);
				xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_PROCS, BAD_CAST s_data);
			}
			processIdCounter+=host->getLoad();
			xmlAddChild(lan_node, node);
		} //if
	} //for iter
	xmlAddChild(topology, lan_node);
	//******END: Rank of root process must be 0 in case of trees****************************

	map<int, void *>::iterator lan_it;
	for (lan_it=LANs.begin(); lan_it != LANs.end(); ++lan_it) {
		void *lan_address=lan_it->second;
		if (lan_address == NULL) continue;

		xmlNodePtr lan_node = xmlNewNode(NULL, BAD_CAST TAAG_XML_LAN);
		sprintf(s_data, "%d", lan_counter++);
		xmlNewProp(lan_node, BAD_CAST TAAG_XML_ATTR_ID, BAD_CAST s_data);

		//*****Hosts*****
		map<int, int>::iterator iter;
		for (iter=lowestProcessIndexOnHosts.begin(); iter != lowestProcessIndexOnHosts.end(); ++iter) {
			if (LANs[iter->first]==lan_address) {
				LANs[iter->first]=NULL;
				lowestProcessIndexOnHosts[iter->first]=processIdCounter;
				HostData *host = hostList->getHostByIndex(iter->first);

				node = xmlNewNode(NULL, BAD_CAST TAAG_XML_HOST);
				sprintf(s_data, "%d", host->getHostId());
				xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_ID, BAD_CAST s_data);
				xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_NAME, BAD_CAST host->getHostname());
				sprintf(s_data, "%d", host->getCPUcount());
				xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_CPUS, BAD_CAST s_data);
				xmlNewChild(node, NULL, BAD_CAST TAAG_XML_DIRECTORY, BAD_CAST dir);
				for (int i=0; i<host->getLoad(); i++) {
					sprintf(s_data, "%d", processIdCounter+i);
					xmlNewChild(node, NULL, BAD_CAST TAAG_XML_H_PROCS, BAD_CAST s_data);
				}
				processIdCounter+=host->getLoad();
				xmlAddChild(lan_node, node);
			} //if
		} //for iter
		xmlAddChild(topology, lan_node);
	} //for lan_it

	//******Groups*****
	for (int id=0; id<nrOfGrps; id++) {
		GroupData *grp = procsDist->getGroupByIndex(id);
		node = xmlNewNode(NULL, BAD_CAST TAAG_XML_GROUP);
		sprintf(s_data, "%d", id);
		xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_ID, BAD_CAST s_data);

		for (int j = 0; j < grp->getNrOfLoadTickets(); j++) {
			//counting the number of processes scheduled to a host
			for (int procs = 0; procs<grp->getMappingTicketByIndex(j)->getNrOfProcs(); procs++) {
				int hostId = grp->getMappingTicketByIndex(j)->getHost()->getHostId();
				sprintf(s_data, "%d", lowestProcessIndexOnHosts[hostId]++);
				xmlNewChild(node, NULL, BAD_CAST TAAG_XML_G_PROCS, BAD_CAST s_data);
			}
		}
		xmlAddChild(graph, node);
	}
	//******Edges*****
	if (procsDist->areEdgesDefined()) {
		for (int k=0; k<nrOfGrps; k++) {
			for (int l=k+1; l<nrOfGrps; l++) {
				if (procsDist->getEdge(k, l)==1) {
					node = xmlNewNode(NULL, BAD_CAST TAAG_XML_EDGE);
					sprintf(s_data, "%d", k);
					xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_FIRST_END, BAD_CAST s_data);
					sprintf(s_data, "%d", l);
					xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_OTHER_END, BAD_CAST s_data);
					xmlAddChild(graph, node);
				}
			}
		}
	}

	return 0;
}

int _add_latency_values(LatencyMatrix *matrix) {
	xmlNodePtr node = NULL;
	char s_data[TAAG_MAX_HOSTNAME_STRING];
	map<int, int>::iterator iter1;

	for (iter1=lowestProcessIndexOnHosts.begin(); iter1 != lowestProcessIndexOnHosts.end(); ++iter1) {
		map<int, int>::const_iterator iter2;
		for (iter2=iter1; iter2 != lowestProcessIndexOnHosts.end(); ++iter2) {
			sprintf(s_data, "%.4f", matrix->getValue(iter1->first, iter2->first)->getValue());
			node = xmlNewChild(topology, NULL, BAD_CAST TAAG_XML_LATENCY, BAD_CAST s_data);
			sprintf(s_data, "%d", iter1->first);
			xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_ROW, BAD_CAST s_data);
			sprintf(s_data, "%d", iter2->first);
			xmlNewProp(node, BAD_CAST TAAG_XML_ATTR_COLUMN, BAD_CAST s_data);
		}
	}
	return 0;
}

int _save_xml_mapping(char *filename) {
	char s_data[TAAG_MAX_FILENAME_STRING];

	xmlAddChild(root_node, graph);
	xmlAddChild(root_node, topology);
	/*
	 * Dumping document to stdio or file
	 */
	sprintf(s_data,"/tmp/%s.xml", appId);
	xmlSaveFormatFileEnc(s_data, doc, "UTF-8", 1);

	/*free the document */
	xmlFreeDoc(doc);

	/*
	 *Free the global variables that may
	 *have been allocated by the parser.
	 */
	xmlCleanupParser();

	/*
	 * this is to debug memory for regression tests
	 */
	xmlMemoryDump();
	strcpy(filename, s_data);
	return 0;
}
